package com.perforce.hws.swagger.asciidoc;

import com.perforce.hws.swagger.asciidoc.models.Definition;
import com.perforce.hws.swagger.asciidoc.models.Route;
import io.swagger.models.HttpMethod;
import io.swagger.models.Operation;
import io.swagger.models.Swagger;
import io.swagger.parser.SwaggerParser;
import org.stringtemplate.v4.ST;
import org.stringtemplate.v4.STGroupFile;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;

/**
 * Creates asciidoc fragments for paths and definitions from the swagger
 * definitions.
 *
 * Initially, we tried using Swagger2Markup, but our platform URL versioning
 * scheme doesn't quite fit the swagger methods. This generator fixes the
 * main issues.
 *
 * We want to be able to include the URL definitions as the bulk of a whole,
 * that includes a few non-platform version URLs, like status and "what versions
 * do you support" methods.
 */
public class GenerateAsciidoc implements HWSSwaggerMethods {

    /**
     * Usage: java ... GenerateAsciidoc source.yaml output_dir
     *
     * @param args
     */
    public static void main(String args[]) throws Exception {
        new GenerateAsciidoc(args[1], args[0]).run();
    }

    private String sourceYaml;
    private String outputDir;

    public GenerateAsciidoc(String outputDir, String sourceYaml) {
        this.outputDir = outputDir;
        this.sourceYaml = sourceYaml;
    }

    public void run() throws Exception {
        if (!Files.exists(Paths.get(outputDir))) {
            Files.createDirectories(Paths.get(outputDir));
        }

        Swagger swagger = new SwaggerParser().read(sourceYaml);

        createDefaultPaths(swagger);
        createAlphaPaths(swagger);
        createDefs(swagger);

        List<Route> routes = convertToRoutes(swagger);
        createJavaApi(routes);
        createJavascriptApi(routes);
        createRubyApi(routes);
        createPhpApi(routes);
        createPythonApi(routes);

        List<Definition> definitions = convertToDefinitions(swagger);
        createJavaModels(definitions);
        createJavascriptModels(definitions);
        createRubyModels(definitions);
        createPhpModels(definitions);
        createPythonModels(definitions);
    }

    private void createDefaultPaths(Swagger swagger) throws IOException {
        STGroupFile pathsGroupFile = new STGroupFile("asciidoc/paths.stg");
        ST pathTemplate = pathsGroupFile.getInstanceOf("format_path");
        pathTemplate.add("swagger", swagger);

        StringBuilder pathSb = new StringBuilder();

        swagger.getPaths().forEach((subPath, path) -> {
            path.getOperationMap().forEach((method, operation) -> {
                if (operation.getTags() == null || !operation.getTags().contains("alpha")) {
                    writePathTemplate(pathTemplate, pathSb, subPath, path, method, operation);
                }
            });
        });

        Files.write(defaultPathsOutputFile(), pathSb.toString().getBytes());
    }

    private void createAlphaPaths(Swagger swagger) throws IOException {
        STGroupFile pathsGroupFile = new STGroupFile("asciidoc/paths.stg");
        ST pathTemplate = pathsGroupFile.getInstanceOf("format_path");
        pathTemplate.add("swagger", swagger);

        StringBuilder pathSb = new StringBuilder();

        swagger.getPaths().forEach((subPath, path) -> {
            path.getOperationMap().forEach((method, operation) -> {
                if (operation.getTags() != null && operation.getTags().contains("alpha")) {
                    writePathTemplate(pathTemplate, pathSb, subPath, path, method, operation);
                }
            });
        });

        Files.write(alphaPathsOutputFile(), pathSb.toString().getBytes());
    }

    private void writePathTemplate(ST pathTemplate, StringBuilder pathSb, String subPath,
                                   io.swagger.models.Path path, HttpMethod method,
                                   Operation operation) {
        pathTemplate.add("subPath", subPath);
        pathTemplate.add("path", path);
        pathTemplate.add("method", method);
        pathTemplate.add("operation", operation);

        pathSb.append(pathTemplate.render());

        pathTemplate.remove("subPath");
        pathTemplate.remove("path");
        pathTemplate.remove("method");
        pathTemplate.remove("operation");
    }

    private void createDefs(Swagger swagger) throws IOException {
        STGroupFile defsGroupFile = new STGroupFile("asciidoc/defs.stg");

        ST defsTemplate = defsGroupFile.getInstanceOf("format_defs");
        defsTemplate.add("swagger", swagger);

        StringBuilder defsSb = new StringBuilder();
        defsSb.append(defsTemplate.render());

        Files.write(definitionsOutputFile(), defsSb.toString().getBytes());
    }

    private void createJavaApi(List<Route> routes) throws IOException {
        STGroupFile javaGroupFile = new STGroupFile("asciidoc/java.stg");

        ST javaTemplate = javaGroupFile.getInstanceOf("format_java_method");

        StringBuilder defaultSb = new StringBuilder();
        StringBuilder alphaSb = new StringBuilder();

        routes.forEach(route -> {
            StringBuilder sb;
            if (route.isAlpha()) {
                sb = alphaSb;
            } else {
                sb = defaultSb;
            }

            javaTemplate.add("route", route);

            sb.append(javaTemplate.render());

            javaTemplate.remove("route");
        });

        Files.write(defaultJavaApiOutputFile(), defaultSb.toString().getBytes());
        Files.write(alphaJavaApiOutputFile(), alphaSb.toString().getBytes());
    }

    private void createJavaModels(List<Definition> definitions) throws IOException {
        STGroupFile javaGroupFile = new STGroupFile("asciidoc/java.stg");

        ST javaTemplate = javaGroupFile.getInstanceOf("java_model");

        StringBuilder sb = new StringBuilder();

        definitions.forEach(d -> {
            javaTemplate.add("model", d);
            sb.append(javaTemplate.render());
            javaTemplate.remove("model");
        });

        Files.write(javaModelOutputFile(), sb.toString().getBytes());
    }

    private void createJavascriptApi(List<Route> routes) throws IOException {
        STGroupFile jsGroupFile = new STGroupFile("asciidoc/javascript.stg");

        ST template = jsGroupFile.getInstanceOf("api_method");

        StringBuilder defaultSb = new StringBuilder();
        StringBuilder alphaSb = new StringBuilder();

        routes.forEach(route -> {
            StringBuilder sb;
            if (route.isAlpha()) {
                sb = alphaSb;
            } else {
                sb = defaultSb;
            }

            template.add("route", route);

            sb.append(template.render());

            template.remove("route");
        });

        Files.write(defaultJsApiOutputFile(), defaultSb.toString().getBytes());
        Files.write(alphaJsApiOutputFile(), alphaSb.toString().getBytes());
    }

    private void createJavascriptModels(List<Definition> definitions) throws IOException {
        STGroupFile javaGroupFile = new STGroupFile("asciidoc/javascript.stg");

        ST javaTemplate = javaGroupFile.getInstanceOf("javascript_model");

        StringBuilder sb = new StringBuilder();

        definitions.forEach(d -> {
            javaTemplate.add("model", d);
            sb.append(javaTemplate.render());
            javaTemplate.remove("model");
        });

        Files.write(javascriptModelOutputFile(), sb.toString().getBytes());
    }

    private void createRubyApi(List<Route> routes) throws IOException {
        STGroupFile javaGroupFile = new STGroupFile("asciidoc/ruby.stg");

        ST javaTemplate = javaGroupFile.getInstanceOf("ruby_method");

        StringBuilder defaultSb = new StringBuilder();
        StringBuilder alphaSb = new StringBuilder();

        routes.forEach(route -> {
            StringBuilder sb;
            if (route.isAlpha()) {
                sb = alphaSb;
            } else {
                sb = defaultSb;
            }

            javaTemplate.add("route", route);

            sb.append(javaTemplate.render());

            javaTemplate.remove("route");
        });

        Files.write(defaultRubyApiOutputFile(), defaultSb.toString().getBytes());
        Files.write(alphaRubyApiOutputFile(), alphaSb.toString().getBytes());
    }

    private void createRubyModels(List<Definition> definitions) throws IOException {
        STGroupFile javaGroupFile = new STGroupFile("asciidoc/ruby.stg");

        ST javaTemplate = javaGroupFile.getInstanceOf("ruby_model");

        StringBuilder sb = new StringBuilder();

        definitions.forEach(d -> {
            javaTemplate.add("model", d);
            sb.append(javaTemplate.render());
            javaTemplate.remove("model");
        });

        Files.write(rubyModelOutputFile(), sb.toString().getBytes());
    }

    private void createPhpApi(List<Route> routes) throws IOException {
        STGroupFile javaGroupFile = new STGroupFile("asciidoc/php.stg");

        ST javaTemplate = javaGroupFile.getInstanceOf("php_method");

        StringBuilder defaultSb = new StringBuilder();
        StringBuilder alphaSb = new StringBuilder();

        routes.forEach(route -> {
            StringBuilder sb;
            if (route.isAlpha()) {
                sb = alphaSb;
            } else {
                sb = defaultSb;
            }

            javaTemplate.add("route", route);

            sb.append(javaTemplate.render());

            javaTemplate.remove("route");
        });

        Files.write(defaultPhpApiOutputFile(), defaultSb.toString().getBytes());
        Files.write(alphaPhpApiOutputFile(), alphaSb.toString().getBytes());
    }

    private void createPhpModels(List<Definition> definitions) throws IOException {
        STGroupFile javaGroupFile = new STGroupFile("asciidoc/php.stg");

        ST javaTemplate = javaGroupFile.getInstanceOf("php_model");

        StringBuilder sb = new StringBuilder();

        definitions.forEach(d -> {
            javaTemplate.add("model", d);
            sb.append(javaTemplate.render());
            javaTemplate.remove("model");
        });

        Files.write(phpModelOutputFile(), sb.toString().getBytes());
    }

    private void createPythonApi(List<Route> routes) throws IOException {
        STGroupFile javaGroupFile = new STGroupFile("asciidoc/python.stg");

        ST javaTemplate = javaGroupFile.getInstanceOf("python_method");

        StringBuilder defaultSb = new StringBuilder();
        StringBuilder alphaSb = new StringBuilder();

        routes.forEach(route -> {
            StringBuilder sb;
            if (route.isAlpha()) {
                sb = alphaSb;
            } else {
                sb = defaultSb;
            }

            javaTemplate.add("route", route);

            sb.append(javaTemplate.render());

            javaTemplate.remove("route");
        });

        Files.write(defaultPythonApiOutputFile(), defaultSb.toString().getBytes());
        Files.write(alphaPythonApiOutputFile(), alphaSb.toString().getBytes());
    }


    private void createPythonModels(List<Definition> definitions) throws IOException {
        STGroupFile javaGroupFile = new STGroupFile("asciidoc/python.stg");

        ST javaTemplate = javaGroupFile.getInstanceOf("python_model");

        StringBuilder sb = new StringBuilder();

        definitions.forEach(d -> {
            javaTemplate.add("model", d);
            sb.append(javaTemplate.render());
            javaTemplate.remove("model");
        });

        Files.write(pythonModelOutputFile(), sb.toString().getBytes());
    }

    private Path defaultPathsOutputFile() {
        return Paths.get(outputDir, "paths-default.adoc");
    }

    private Path alphaPathsOutputFile() {
        return Paths.get(outputDir, "paths-alpha.adoc");
    }

    private Path definitionsOutputFile() {
        return Paths.get(outputDir, "definitions.adoc");
    }

    private Path defaultJavaApiOutputFile() {
        return Paths.get(outputDir, "java-default.adoc");
    }

    private Path alphaJavaApiOutputFile() {
        return Paths.get(outputDir, "java-alpha.adoc");
    }

    private Path javaModelOutputFile() {
        return Paths.get(outputDir, "java-models.adoc");
    }

    private Path defaultJsApiOutputFile() {
        return Paths.get(outputDir, "js-default.adoc");
    }

    private Path alphaJsApiOutputFile() {
        return Paths.get(outputDir, "js-alpha.adoc");
    }

    private Path javascriptModelOutputFile() {
        return Paths.get(outputDir, "javascript-models.adoc");
    }

    private Path defaultRubyApiOutputFile() {
        return Paths.get(outputDir, "ruby-default.adoc");
    }

    private Path alphaRubyApiOutputFile() {
        return Paths.get(outputDir, "ruby-alpha.adoc");
    }

    private Path rubyModelOutputFile() {
        return Paths.get(outputDir, "ruby-models.adoc");
    }

    private Path defaultPhpApiOutputFile() {
        return Paths.get(outputDir, "php-default.adoc");
    }

    private Path alphaPhpApiOutputFile() {
        return Paths.get(outputDir, "php-alpha.adoc");
    }

    private Path phpModelOutputFile() {
        return Paths.get(outputDir, "php-models.adoc");
    }

    private Path defaultPythonApiOutputFile() {
        return Paths.get(outputDir, "python-default.adoc");
    }

    private Path alphaPythonApiOutputFile() {
        return Paths.get(outputDir, "python-alpha.adoc");
    }

    private Path pythonModelOutputFile() {
        return Paths.get(outputDir, "python-models.adoc");
    }
}
